package se.gawell.setrace.spring;

/*
 *     This file is part of SeTrace.
 *
 *   SeTrace is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   SeTrace is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SeTrace.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

import java.util.List;

import net.sf.cglib.proxy.Callback;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.NoOp;

import org.springframework.aop.TargetSource;
import org.springframework.aop.framework.autoproxy.AbstractAutoProxyCreator;
import org.springframework.util.PatternMatchUtils;

import se.gawell.setrace.logging.Logger;
import se.gawell.setrace.logging.LoggerFactory;

/**
 * Auto proxy creator for creating proxies used for logging method invocations.
 * 
 * Use this as an AbstractAutoProxyCreator.
 * 
 * @author Niklas Gawell
 *
 */
public class ExcludeMappedNameBeanNameAutoProxyCreator extends
		AbstractAutoProxyCreator {

	private final LoggerFactory loggerFactory;
	private final List<String> excludedNames;

	public ExcludeMappedNameBeanNameAutoProxyCreator(LoggerFactory loggerFactory, List<String> excludedNames) {
		super();
		this.loggerFactory = loggerFactory;
		this.excludedNames = excludedNames;
	}

	@Override
	protected Object[] getAdvicesAndAdvisorsForBean(Class<?> beanClass,
			String beanName, TargetSource targetSource) {
		if (classShouldBeExcludedSinceItIsNotProxable(beanClass, beanName)) {
			return doNotProxy(beanName);
		}
		if (classIsAnnotatedWithExcludeAnnotation(beanClass, beanName)) {
			return doNotProxy(beanName);
		}
		for (String nameToExclude : excludedNames) {
			if (beanNameMatchExclude(beanClass.getName(), nameToExclude)) {
				return doNotProxy(beanName);
			}
		}
		logProxyCreation(beanName);
		return PROXY_WITHOUT_ADDITIONAL_INTERCEPTORS;
	}

	private Object[] doNotProxy(String beanName) {
		logDoNotProxy(beanName);
		return DO_NOT_PROXY;
	}

	private boolean classIsAnnotatedWithExcludeAnnotation(Class<?> beanClass, String beanName) {
		boolean annotationPresent = beanClass.isAnnotationPresent(ExcludeFromTrace.class);
		if (annotationPresent) {
			getLogger().debug("{} is annotated with {} and will not be proxied", beanName, ExcludeFromTrace.class.getSimpleName());
		}
		return annotationPresent;
	}

	private void logDoNotProxy(String beanName) {
		Logger loggerForClass = getLogger();
		loggerForClass.debug("Not adding method execution logger to bean {0}.", beanName);
	}

	private void logProxyCreation(String beanName) {
		Logger loggerForClass = getLogger();
		loggerForClass.debug("Adding method execution logger to bean {0}.", beanName);
	}

	private boolean classShouldBeExcludedSinceItIsNotProxable(Class<?> beanClass, String beanName) {
		Logger loggerForClass = getLogger();
		if (beanClass.getInterfaces().length > 0) {
			loggerForClass.debug("{} has interface and is thus proxable", beanName);
			return false;
		}
		try {
			Callback callback = NoOp.INSTANCE;
			Enhancer.create(beanClass, callback);
		} catch (RuntimeException e) {
			loggerForClass
					.debug("Cannot proxy bean {} due to {}.", beanName, e.toString());
			return true;
		}
		return false;
	}

	private Logger getLogger() {
		return loggerFactory.getLoggerForClass(getClass());
	}

	private boolean beanNameMatchExclude(String beanName, String nameToExclude) {
		return PatternMatchUtils.simpleMatch(nameToExclude, beanName);
	}

}
